
#ifndef __KERNEL__
 #define __KERNEL__
#endif

#ifndef MODULE
 #define MODULE
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/pci.h>

#include "dpci16in16out.h"
#include "sysdep.h"   

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>
#endif

/* I defined some macro for device_open function */
#define TYPE(dev) (MINOR(dev) >> 4) 
#define NUM(dev) (MINOR(dev) & 0x0f)

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#ifndef DECISION_VENDOR_ID
#define DECISION_VENDER_ID 0x6666
#endif

#ifndef DECISION_DEVICE_ID
#define DECISION_DEVICE_ID 0x1025
#endif


static int cards;
static char rev[2];
int Base_Addr[MAXCARD];

/*---------------------------------------------------------------------*/
int dpci16in16out_open(struct inode *inoder,struct file *flip)
{
/* int type= TYPE(inoder->i_rdev);*/

 int num=NUM(inoder->i_rdev);
 if(num>=cards) return -ENODEV;
 MOD_INC_USE_COUNT;
 return 0;
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int dpci16in16out_release(struct inode *inoder,struct file *flip)
#else
static void dpci16in16out_release(struct inode *inoder,struct file *flip)
#endif
{
  MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif    
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t dpci16in16out_read(struct file *flip,char *buf,size_t count,loff_t *ptr)
#else
static int dpci16in16out_read(struct inode *inode,struct file *flip, char *buf,count_t count)
#endif
{
 int countt=0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  int num=MINOR(flip->f_dentry->d_inode->i_rdev);
  if(ptr!=&flip->f_pos)  /* can't seek on this device */
   return -ESPIPE;
#else
  int num=NUM(inode->i_rdev);
#endif 
 
 if(num>=cards)
   return -EINVAL;
 buf[0]=inb(Base_Addr[num]);
 countt=1;
           
 return countt;
}
/*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t dpci16in16out_write(struct file *flip,const char *buf,size_t count,loff_t *ptr)
#else
static int dpci16in16out_write(struct inde *inode,struct file *flip,char *buf,count_t count)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)	
 if(ptr!=&flip->f_pos)  /* can't seek on this device */
  return -ESPIPE;
#endif
 
 return 0;       /* do nothing */
}
/*---------------------------------------------------------------------*/
int Myread(int num,char *buf,int count)
{
 int countt;
 buf[0]=inb(Base_Addr[num]);
 buf[1]=inb(Base_Addr[num]+1);
 buf[0]= ~buf[0];
 buf[1]= ~buf[1];
 countt=2;
 return countt;
}
/*---------------------------------------------------------------------*/
void Mywrite(int num,char *buf,int count)
{
 outb(buf[0],Base_Addr[num]);
 outb(buf[1],Base_Addr[num]+1);
}
/*---------------------------------------------------------------------*/
int dpci16in16out_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
  char param[4];
  char *temp=(char *)ioctl_param;
  int num=NUM(inode->i_rdev);
  int data;
  switch (ioctl_num) 
   {
    case DPCI16IN16OUT_WRITE_PORT:
      data=(int)temp;
      param[0]=(char)(data&0xff);
      param[1]=(char)((data&0xff00)>>8);
      Mywrite(num,param,2);
      return data;
      break;

    case DPCI16IN16OUT_READ_PORT:
      Myread(num,param,2);
      put_user(param[0],temp);
      temp++;
      put_user(param[1],temp);
      break;

     default:
             break;
  }
  return 0;
}
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
static struct file_operations dpci16in16out_fops={
read: dpci16in16out_read,
write: dpci16in16out_write,
ioctl: dpci16in16out_ioctl ,
open: dpci16in16out_open,
release: dpci16in16out_release,
};
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
MODULE_AUTHOR("Delta");
MODULE_DESCRIPTION("Decision PCI 16in/16out Card Driver");
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/

static int __init dpci16in16out_init(void)
{
 int index=0;
 struct pci_dev *dev = NULL;
 static int dev_count=0;
 int i; 
 dev_count++;
 if(dev_count>1)
  {
   printk(KERN_ERR"This drive only can open one time.\n");
   return -ENODEV;
  }

#ifdef CONFIG_PCI
 if (!pci_present())   /* No PCI bus in this machine! */
#endif
 return -ENODEV;

 cards=0;
 while (index < MAXCARD && 
        (dev = pci_find_device(DECISION_VENDER_ID, DECISION_DEVICE_ID,dev))) 
 {
   Base_Addr[index]=pci_resource_start(dev,3) & PCI_BASE_ADDRESS_IO_MASK;   
   printk(KERN_ERR "PCI 16in/16out card:I got a base 0x%x\n",Base_Addr[index]); 
   if(check_region(Base_Addr[index],REGIONLENGTH)>=0)
    request_region(Base_Addr[index],REGIONLENGTH,DEVICE_NAME);
   index++; 
  }

  if((dev==NULL)&&(index<1))
   {
    printk("\nI can't find any Decision PCI 16in/16out card here!!\n");
    return -ENODEV;
   }
 printk("\nPCI 16in/16out card: Total %d cards that I catched!\n",index);
 cards=index;
 index=register_chrdev(dpci16in16out_major,DEVICE_NAME,&dpci16in16out_fops);
 if(index<0)
  {
  for(i=0;i<cards;i++)
   release_region(Base_Addr[i],REGIONLENGTH);
  return index;
  }
 rev[0]=0;
 printk("PCI 16in/16out card:Register Successful\n");  
 return 0;

}
/*---------------------------------------------------------------------*/
static void __exit dpci16in16out_exit(void)
{
  int i;
  for(i=0;i<cards;i++)
   release_region(Base_Addr[i],REGIONLENGTH); 
  unregister_chrdev(dpci16in16out_major,DEVICE_NAME);
  printk("PCI 16in/16out card:Unregister Successful!!\n");  
}
/*---------------------------------------------------------------------*/
module_init(dpci16in16out_init);
module_exit(dpci16in16out_exit);
